/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #   - DIDELOT Sylvain didelot.sylvain@gmail.com                        # */
/* #                                                                      # */
/* ######################################################################## */

#ifdef MPC_USE_INFINIBAND

#include <sctk.h>
#include "sctk_ib_config.h"
#include "sctk_ib_comp_rc_sr.h"
#include "sctk_ib_comp_rc_rdma.h"

/* Maximum message size for each channels (in fact this in
 * not really a limit...). Here is the algorithm to
 * select which channel use:
 * if x < IBV_EAGER_LIMIT -> eager msg
 * if x < IBV_FRAG_EAGER_LIMIT -> frag msg (into several eager buffers)
 * if x > IBV_FRAG_EAGER_LIMIT -> rendezvous msg */
#define IBV_EAGER_LIMIT ( 12 * 1024 )
#define IBV_FRAG_EAGER_LIMIT ( 256 * 1024)
/* Number of allowed pending Work Queue Elements
 * for each QP */
#define IBV_QP_TX_DEPTH     4000
/* We don't need recv WQE when using SRQ.
 * This variable must be set to 0 */
#define IBV_QP_RX_DEPTH     0
/* Many CQE. In memory, it represents about
 * 1.22Mb for 40000 entries */
#define IBV_CQ_DEPTH        40000
#define IBV_MAX_SG_SQ       8
#define IBV_MAX_SG_RQ       8
#define IBV_MAX_INLINE      128

/* Maximum number of buffers to allocate during the
 * initialization step */
#define IBV_MAX_IBUFS         10000
#define IBV_MAX_SRQ_IBUFS     8000
/* Minimum number of free recv buffer before
 * posting of new buffers. This thread is  activated
 * once a recv buffer is freed. If IBV_SRQ_CREDIT_LIMIT ==
 * IBV_MAX_SRQ_IBUFS, receive buffers are re-post every-time
 * they are consumned */
#define IBV_SRQ_CREDIT_LIMIT  8000
/* Minimum number of free recv buffer before
 * the activation of the asynchronous
 * thread (if this thread is activated too much times,
 * the performance can be decreased) */
#define IBV_SRQ_CREDIT_THREAD_LIMIT  400

/* for PN */
#if 0
#define IBV_MAX_IBUFS         400
#define IBV_MAX_SRQ_IBUFS     300 /*  > 300 */
#define IBV_SRQ_CREDIT_LIMIT  200 /* >=300 */
#define IBV_SRQ_CREDIT_THREAD_LIMIT  200
#endif

/* Number of new buffers allocated when
 * no more buffers are available */
#define IBV_SIZE_IBUFS_CHUNKS 400

#define IBV_WC_IN_NUMBER    1
#define IBV_WC_OUT_NUMBER   1

/* Numer of MMU entries allocated during
 * the MPC initialization */
#define IBV_MAX_MR          100
/* Number of new MMU allocated when
 * no more MMU entries are available */
#define IBV_SIZE_MR_CHUNKS  200

#define IBV_ADM_PORT        1

#define IBV_RDMA_DEPTH       4
#define IBV_RDMA_DEST_DEPTH  4

#define IBV_NO_MEMORY_LIMITATION  1
/* Verbosity level (some infos can appears on
 * the terminal during runtime: new ibufs allocated,
 * new MMu entries allocated, etc...) */
#define IBV_VERBOSE_LEVEL         0

#define IBV_ADAPTIVE_POLLING      1

/* global values */
unsigned int  ibv_eager_limit  = IBV_EAGER_LIMIT;
unsigned int  ibv_frag_eager_limit  = IBV_FRAG_EAGER_LIMIT;
unsigned int  ibv_qp_tx_depth      = IBV_QP_TX_DEPTH;
unsigned int  ibv_qp_rx_depth      = IBV_QP_RX_DEPTH;
unsigned int  ibv_cq_depth         = IBV_CQ_DEPTH;
unsigned int  ibv_max_sg_sq        = IBV_MAX_SG_SQ;
unsigned int  ibv_max_sg_rq        = IBV_MAX_SG_RQ;
unsigned int  ibv_max_inline       = IBV_MAX_INLINE;
unsigned int  ibv_max_ibufs        = IBV_MAX_IBUFS;
unsigned int  ibv_size_ibufs_chunk = IBV_SIZE_IBUFS_CHUNKS;
unsigned int  ibv_max_srq_ibufs    = IBV_MAX_SRQ_IBUFS;
unsigned int  ibv_srq_credit_limit = IBV_SRQ_CREDIT_LIMIT;
unsigned int  ibv_srq_credit_thread_limit = IBV_SRQ_CREDIT_THREAD_LIMIT;
unsigned int  ibv_rdvz_protocol    = IBV_RDVZ_READ_PROTOCOL;
unsigned int  ibv_no_memory_limitation    = IBV_NO_MEMORY_LIMITATION;
unsigned int  ibv_verbose_level    = IBV_VERBOSE_LEVEL;

unsigned int  ibv_wc_in_number     = IBV_WC_IN_NUMBER;
unsigned int  ibv_wc_out_number    = IBV_WC_OUT_NUMBER;
unsigned int  ibv_max_mr           = IBV_MAX_MR;
unsigned int  ibv_size_mr_chunk    = IBV_SIZE_MR_CHUNKS;
unsigned int  ibv_adm_port         = IBV_ADM_PORT;
unsigned int  ibv_rdma_depth       = IBV_RDMA_DEPTH;
unsigned int  ibv_rdma_dest_depth  = IBV_RDMA_DEST_DEPTH;
unsigned int  ibv_adaptive_polling = IBV_ADAPTIVE_POLLING;


void sctk_net_ibv_config_check()
{
//  ibv_eager_limit += (RC_SR_HEADER_SIZE + sizeof(sctk_thread_ptp_message_t));
//    ibv_eager_limit += (RC_SR_HEADER_SIZE);

//   ibv_eager_limit = ( ( ibv_eager_limit + 4095 ) & ( ~4095 ) );

  if (ibv_eager_limit < (RC_SR_HEADER_SIZE + sizeof(sctk_net_ibv_rc_rdma_request_t)))
  {
      goto error;
  }
  if (ibv_eager_limit < (RC_SR_HEADER_SIZE + sizeof(sctk_net_ibv_rc_rdma_coll_request_t)))
  {
      goto error;
  }
  if (ibv_eager_limit < (RC_SR_HEADER_SIZE + sizeof(sctk_net_ibv_rc_rdma_request_ack_t)))
  {
      goto error;
  }
  if (ibv_eager_limit < (RC_SR_HEADER_SIZE + sizeof(sctk_net_ibv_rc_rdma_done_t)))
  {
      goto error;
  }

  return;

error:
  sctk_error("Wrong IB configuration");
  sctk_abort();
  return;
}

void sctk_net_ibv_config_print()
{
  if (sctk_process_rank == 0)
  {
    fprintf(stderr,
        "############# IB CONFIGURATION #############\n"
        "ibv_eager_limit      = %d\n"
        "ibv_frag_eager_limit = %d\n"
        "ibv_qp_tx_depth      = %d\n"
        "ibv_qp_rx_depth      = %d\n"
        "ibv_max_sg_sq        = %d\n"
        "ibv_max_sg_rq        = %d\n"
        "ibv_max_inline       = %d\n"
        "ibv_max_ibufs        = %d\n"
        "ibv_max_srq_ibufs    = %d\n"
        "ibv_size_ibufs_chunk = %d\n"
        "ibv_srq_credit_limit = %d\n"
        "ibv_srq_credit_thread_limit = %d\n"
        "ibv_rdvz_protocol    = %d\n"
        "ibv_wc_in_number     = %d\n"
        "ibv_wc_out_number    = %d\n"
        "ibv_max_mr           = %d\n"
        "ibv_adm_port         = %d\n"
        "ibv_rdma_depth       = %d\n"
        "ibv_rdma_dest_depth  = %d\n"
        "ibv_adaptive_polling = %d\n"
        "############# IB CONFIGURATION #############\n",
        ibv_eager_limit,
        ibv_frag_eager_limit,
        ibv_qp_tx_depth,
        ibv_qp_rx_depth,
        ibv_max_sg_sq,
        ibv_max_sg_rq,
        ibv_max_inline,
        ibv_max_ibufs,
        ibv_max_srq_ibufs,
        ibv_size_ibufs_chunk,
        ibv_srq_credit_limit,
        ibv_srq_credit_thread_limit,
        ibv_rdvz_protocol,
        ibv_wc_in_number,
        ibv_wc_out_number,
        ibv_max_mr,
        ibv_adm_port,
        ibv_rdma_depth,
        ibv_rdma_dest_depth,
        ibv_adaptive_polling);
  }
}

void sctk_net_ibv_config_init()
{
  char* value;

  if ( (value = getenv("MPC_IBV_EAGER_LIMIT")) != NULL )
    ibv_eager_limit = atoi(value);

  if ( (value = getenv("MPC_IBV_FRAG_EAGER_LIMIT")) != NULL )
    ibv_frag_eager_limit = atoi(value);

  if ( (value = getenv("MPC_IBV_QP_TX_DEPTH")) != NULL )
    ibv_qp_tx_depth = atoi(value);

  if ( (value = getenv("MPC_IBV_QP_RX_DEPTH")) != NULL )
    ibv_qp_rx_depth = atoi(value);

  if ( (value = getenv("MPC_IBV_MAX_SG_SQ")) != NULL )
    ibv_max_sg_sq = atoi(value);

  if ( (value = getenv("MPC_IBV_MAX_RG_SQ")) != NULL )
    ibv_max_sg_rq = atoi(value);

  if ( (value = getenv("MPC_IBV_MAX_INLINE")) != NULL )
    ibv_max_inline = atoi(value);

  if ( (value = getenv("MPC_IBV_MAX_IBUFS")) != NULL )
    ibv_max_ibufs = atoi(value);

  if ( (value = getenv("MPC_IBV_MAX_SRQ_IBUFS")) != NULL )
    ibv_max_srq_ibufs = atoi(value);

  if ( (value = getenv("MPC_IBV_SRQ_CREDIT_LIMIT")) != NULL )
    ibv_srq_credit_limit = atoi(value);

  if ( (value = getenv("MPC_IBV_SRQ_CREDIT_THREAD_LIMIT")) != NULL )
    ibv_srq_credit_thread_limit = atoi(value);

  if ( (value = getenv("MPC_IBV_RDVZ_WRITE_PROTOCOL")) != NULL )
    ibv_rdvz_protocol = IBV_RDVZ_WRITE_PROTOCOL;
  if ( (value = getenv("MPC_IBV_RDVZ_READ_PROTOCOL")) != NULL )
    ibv_rdvz_protocol = IBV_RDVZ_READ_PROTOCOL;

  if ( (value = getenv("MPC_IBV_WC_IN_NUMBER")) != NULL )
    ibv_wc_in_number = atoi(value);

  if ( (value = getenv("MPC_IBV_WC_OUT_NUMBER")) != NULL )
    ibv_wc_out_number = atoi(value);

  if ( (value = getenv("MPC_IBV_MAX_MR")) != NULL )
    ibv_max_mr = atoi(value);

  if ( (value = getenv("MPC_IBV_ADM_PORT")) != NULL )
    ibv_adm_port = atoi(value);

  if ( (value = getenv("MPC_IBV_RDMA_DEPTH")) != NULL )
    ibv_rdma_depth = atoi(value);

  if ( (value = getenv("MPC_IBV_RDMA_DEST_DEPTH")) != NULL )
    ibv_rdma_dest_depth = atoi(value);

  if ( (value = getenv("MPC_IBV_QP_TX_DEPTH")) != NULL )
    ibv_qp_tx_depth = atoi(value);

  if ( (value = getenv("MPC_IBV_SIZE_IBUFS_CHUNK")) != NULL )
    ibv_size_ibufs_chunk = atoi(value);

  if ( (value = getenv("MPC_IBV_ADAPTIVE_POLLING")) != NULL )
  {
    ibv_adaptive_polling = atoi(value);
  }
  /*
   * Check if the variables are well set and print them
   * */
  sctk_net_ibv_config_check();
#if IBV_VERBOSE_LEVEL > 0
  sctk_net_ibv_config_print();
#endif
}
#endif
